from __future__ import annotations

import os
import subprocess

import numpy as np

import audmath


MAX_CHANNELS = {
    "wav": 65535,
    "ogg": 255,
    "mp3": 2,
    "flac": 8,
}
r"""Maximum number of channels per format."""

SNDFORMATS = ["wav", "flac", "mp3", "ogg"]
r"""File formats handled by soundfile"""


def binary_missing_error(binary: str) -> Exception:
    r"""Missing binary error message.

    Args:
        binary: name of binary

    Returns:
        error message

    """
    return FileNotFoundError(
        f"{binary} cannot be found.\n"
        "Please make sure it is installed.\n"
        "For further instructions visit: "
        "https://audeering.github.io/audiofile/installation.html"
    )


def broken_file_error(file: str) -> Exception:
    r"""Broken file error message.

    If we encounter a broken file,
    we raise the same error for non SND files
    as soundfile does for SND files

    Args:
        file: file name

    Returns:
        error message

    """
    return RuntimeError(
        f"Error opening {file}: File contains data in an unknown format."
    )


def duration_in_seconds(
    duration: float | int | str | np.timedelta64,
    sampling_rate: float | int,
) -> np.floating:
    r"""Duration in seconds.

    This mirrors the behavior of audmath.duration_in_seconds()
    but handles only string values without unit,
    e.g. ``'2000'`` as representing samples.

    """
    if isinstance(duration, str):
        duration = audmath.duration_in_seconds(duration, sampling_rate)
    else:
        duration = audmath.duration_in_seconds(duration)
    return duration


def file_extension(path):
    """Lower case file extension."""
    # We are not using `audeer.file_extension()` here,
    # to save another call to `audeer.safe_path()`
    return os.path.splitext(path)[-1][1:].lower()


def run(shell_command):
    """Return the output of a shell command provided as string."""
    out = subprocess.check_output(shell_command, stderr=subprocess.STDOUT)
    try:
        return out.split()[0]
    except IndexError:
        return ""


def run_ffmpeg(infile, outfile, offset, duration, sampling_rate):
    """Convert audio file to WAV file."""
    cmd = ["ffmpeg", "-ss", str(offset), "-i", infile, outfile]
    if duration:
        cmd.insert(-1, "-t")
        cmd.insert(-1, str(duration))
    if sampling_rate:
        cmd.insert(-1, "-ar")
        cmd.insert(-1, str(sampling_rate))
    run(cmd)


def run_sox(infile, outfile, offset, duration, sampling_rate):
    """Convert audio file to WAV file."""
    cmd = ["sox", infile, outfile, "trim", str(offset)]
    if duration:
        cmd.append(str(duration))
    if sampling_rate:
        cmd += ["rate", str(sampling_rate)]
    run(cmd)



from datetime import datetime
import os
import shutil

import toml

import audeer

from audiofile.core.conftest import create_audio_files


config = toml.load(audeer.path("..", "pyproject.toml"))


# Project -----------------------------------------------------------------

project = config["project"]["name"]
copyright = f"2018-{datetime.now().year} audEERING GmbH"
author = ", ".join(author["name"] for author in config["project"]["authors"])
version = audeer.git_repo_version()
title = "Documentation"


# General -----------------------------------------------------------------

master_doc = "index"
extensions = []
source_suffix = ".rst"
exclude_patterns = [
    "build",
    "tests",
    "Thumbs.db",
    ".DS_Store",
    "api-src",
]
templates_path = ["_templates"]
pygments_style = None
extensions = [
    "jupyter_sphinx",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",  # support for Google-style docstrings
    "sphinx_autodoc_typehints",
    "sphinx_copybutton",
    "sphinx.ext.viewcode",
    "sphinx.ext.intersphinx",
    "matplotlib.sphinxext.plot_directive",  # include resulting figures in doc
]

intersphinx_mapping = {
    "audmath": ("https://audeering.github.io/audmath/", None),
    "audresample": ("https://audeering.github.io/audresample/", None),
    "python": ("https://docs.python.org/3/", None),
    "numpy": ("https://docs.scipy.org/doc/numpy/", None),
    "soundfile": ("https://python-soundfile.readthedocs.io/en/latest/", None),
}

# Ignore package dependencies during building the docs
autodoc_mock_imports = ["soundfile"]

copybutton_prompt_text = r">>> |\.\.\. |\$ "
copybutton_prompt_is_regexp = True

# Plotting
plot_include_source = True
plot_html_show_source_link = False
plot_html_show_formats = False
plot_pre_code = ""
plot_rcparams = {
    "figure.figsize": "5, 3.8",  # inch
}
plot_formats = ["png"]

# Disable auto-generation of TOC entries in the API
# https://github.com/sphinx-doc/sphinx/issues/6316
toc_object_entries = False


# HTML --------------------------------------------------------------------

html_theme = "sphinx_audeering_theme"
html_theme_options = {
    "display_version": True,
    "footer_links": False,
    "logo_only": False,
}
html_context = {
    "display_github": True,
}
html_title = title


# Copy API (sub-)module RST files to docs/api/ folder ---------------------
audeer.rmdir("api")
audeer.mkdir("api")
api_src_files = audeer.list_file_names("api-src")
api_dst_files = [
    audeer.path("api", os.path.basename(src_file)) for src_file in api_src_files
]
for src_file, dst_file in zip(api_src_files, api_dst_files):
    shutil.copyfile(src_file, dst_file)


# Create audio files ------------------------------------------------------
create_audio_files("api")



"""Read, write, and get information about audio files."""

from audiofile.core.info import bit_depth
from audiofile.core.info import channels
from audiofile.core.info import duration
from audiofile.core.info import has_video
from audiofile.core.info import samples
from audiofile.core.info import sampling_rate
from audiofile.core.io import convert_to_wav
from audiofile.core.io import read
from audiofile.core.io import write


# Discourage from audeer import *
__all__ = []


# Dynamically get the version of the installed module
try:
    import importlib.metadata

    __version__ = importlib.metadata.version(__name__)
except Exception:  # pragma: no cover
    importlib = None  # pragma: no cover
finally:
    del importlib



"""Read, write, and get information about audio files."""

from __future__ import annotations

import os
import subprocess
import tempfile

import soundfile

import audeer

from audiofile.core.io import convert_to_wav
from audiofile.core.utils import SNDFORMATS
from audiofile.core.utils import binary_missing_error
from audiofile.core.utils import broken_file_error
from audiofile.core.utils import file_extension
from audiofile.core.utils import run


def bit_depth(file: str) -> int | None:
    r"""Bit depth of audio file.

    For lossy audio files,
    ``None`` is returned as they have a varying bit depth.

    Args:
        file: file name of input audio file

    Returns:
        bit depth of audio file

    Raises:
        RuntimeError: if ``file`` is missing,
            broken or format is not supported

    Examples:
        >>> bit_depth("stereo.wav")
        16

    """
    file = audeer.safe_path(file)
    file_type = file_extension(file)
    if file_type == "wav":
        precision_mapping = {
            "PCM_16": 16,
            "PCM_24": 24,
            "PCM_32": 32,
            "PCM_U8": 8,
            "FLOAT": 32,
            "DOUBLE": 64,
            "ULAW": 8,
            "ALAW": 8,
            "IMA_ADPCM": 4,
            "MS_ADPCM": 4,
            "GSM610": 16,  # not sure if this could be variable?
            "G721_32": 4,  # not sure if correct
        }
    elif file_type == "flac":
        precision_mapping = {
            "PCM_16": 16,
            "PCM_24": 24,
            "PCM_32": 32,
            "PCM_S8": 8,
        }
    if file_extension(file) in ["wav", "flac"]:
        depth = precision_mapping[soundfile.info(file).subtype]
    else:
        depth = None

    return depth


def channels(file: str) -> int:
    """Number of channels in audio file.

    Args:
        file: file name of input audio file

    Returns:
        number of channels in audio file

    Raises:
        FileNotFoundError: if mediainfo binary is needed,
            but cannot be found
        RuntimeError: if ``file`` is missing,
            broken or format is not supported

    Examples:
        >>> channels("stereo.wav")
        2

    """
    file = audeer.safe_path(file)
    if file_extension(file) in SNDFORMATS:
        return soundfile.info(file).channels
    else:
        try:
            cmd = ["soxi", "-c", file]
            return int(run(cmd))
        except (FileNotFoundError, subprocess.CalledProcessError):
            # For MP4 stored and returned number of channels can be different
            cmd1 = ["mediainfo", "--Inform=Audio;%Channel(s)_Original%", file]
            cmd2 = ["mediainfo", "--Inform=Audio;%Channel(s)%", file]
            try:
                return int(run(cmd1))
            except FileNotFoundError:
                raise binary_missing_error("mediainfo")
            except (ValueError, subprocess.CalledProcessError):
                try:
                    return int(run(cmd2))
                except (ValueError, subprocess.CalledProcessError):
                    raise broken_file_error(file)


def duration(file: str, sloppy=False) -> float:
    """Duration in seconds of audio file.

    The default behavior (``sloppy=False``)
    ensures
    the duration in seconds
    matches the one in samples.
    To achieve this it first decodes files to WAV
    if needed, e.g. MP4 files.
    If you have different decoders
    on different machines,
    results might differ.

    The case ``sloppy=True`` returns the duration
    as reported in the header of the audio file.
    This is faster,
    but might still return different results
    on different machines
    as it depends on the installed software.
    If no duration information is provided in the header
    it will fall back to ``sloppy=False``.

    Args:
        file: file name of input audio file
        sloppy: if ``True`` report duration
            as stored in the header

    Returns:
        duration in seconds of audio file

    Raises:
        FileNotFoundError: if ffmpeg or mediainfo binary is needed,
            but cannot be found
        RuntimeError: if ``file`` is missing,
            broken or format is not supported

    Examples:
        >>> duration("stereo.wav")
        1.5

    """
    file = audeer.safe_path(file)
    if file_extension(file) in SNDFORMATS:
        return soundfile.info(file).duration

    if sloppy:
        try:
            cmd = ["soxi", "-D", file]
            duration = float(run(cmd))
        except (FileNotFoundError, subprocess.CalledProcessError):
            try:
                cmd = ["mediainfo", "--Inform=Audio;%Duration%", file]
                duration = run(cmd)
                if duration:
                    # Convert to seconds, as mediainfo returns milliseconds
                    duration = float(duration) / 1000
            except FileNotFoundError:
                raise binary_missing_error("mediainfo")
            # Behavior for broken files is different on Windows
            # where no error is raised,
            # but an empty duration is returned.
            # The error under Windows is then raised
            # when calling 'samples(file)'
            except subprocess.CalledProcessError:  # pragma: nocover
                raise broken_file_error(file)
        if duration:
            return duration

    return samples(file) / sampling_rate(file)


def has_video(file: str) -> bool:
    """If file contains video.

    For WAV, FLAC, MP3, or OGG files
    ``False`` is returned.
    All other files are probed by mediainfo.

    Args:
        file: file name of input file

    Returns:
        ``True`` if file contains video data

    Raises:
        FileNotFoundError: if mediainfo binary is needed,
            but cannot be found
        RuntimeError: if ``file`` is missing
            and does not end with ``"wav"``, ``"flac"``, ``"mp3"``, ``"ogg"``

    Examples:
        >>> has_video("stereo.wav")
        False

    """
    if file_extension(file) in SNDFORMATS:
        return False
    else:
        try:
            path = audeer.path(file)
            if not os.path.exists(path):
                raise RuntimeError(f"'{file}' does not exist.")
            cmd = ["mediainfo", "--Inform=Video;%Format%", path]
            video_format = run(cmd)
            if len(video_format) > 0:
                return True
            else:
                return False
        except FileNotFoundError:
            raise binary_missing_error("mediainfo")


def samples(file: str) -> int:
    """Number of samples in audio file.

    Audio files that are not WAV, FLAC, MP3, or OGG
    are first converted to WAV,
    before counting the samples.

    Args:
        file: file name of input audio file

    Returns:
        number of samples in audio file

    Raises:
        FileNotFoundError: if ffmpeg binary is needed,
            but cannot be found
        RuntimeError: if ``file`` is missing,
            broken or format is not supported

    Examples:
        >>> samples("stereo.wav")
        12000

    """

    def samples_as_int(file):
        return int(soundfile.info(file).duration * soundfile.info(file).samplerate)

    file = audeer.safe_path(file)
    if file_extension(file) in SNDFORMATS:
        return samples_as_int(file)
    else:
        # Always convert to WAV for non SNDFORMATS
        with tempfile.TemporaryDirectory(prefix="audiofile") as tmpdir:
            tmpfile = os.path.join(tmpdir, "tmp.wav")
            convert_to_wav(file, tmpfile)
            return samples_as_int(tmpfile)


def sampling_rate(file: str) -> int:
    """Sampling rate of audio file.

    Args:
        file: file name of input audio file

    Returns:
        sampling rate of audio file

    Raises:
        FileNotFoundError: if mediainfo binary is needed,
            but cannot be found
        RuntimeError: if ``file`` is missing,
            broken or format is not supported

    Examples:
        >>> sampling_rate("stereo.wav")
        8000

    """
    file = audeer.safe_path(file)
    if file_extension(file) in SNDFORMATS:
        return soundfile.info(file).samplerate
    else:
        try:
            cmd = ["soxi", "-r", file]
            return int(run(cmd))
        except (FileNotFoundError, subprocess.CalledProcessError):
            try:
                cmd = ["mediainfo", "--Inform=Audio;%SamplingRate%", file]
                sampling_rate = run(cmd)
                if sampling_rate:
                    return int(sampling_rate)
                else:
                    # Raise CalledProcessError
                    # to align coverage under Windows and Linux
                    raise subprocess.CalledProcessError(-2, cmd)
            except FileNotFoundError:
                raise binary_missing_error("mediainfo")
            except subprocess.CalledProcessError:
                raise broken_file_error(file)



import os

import matplotlib
import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns


matplotlib.use("Agg")
DEVNULL = open(os.devnull, "w")


class DFWriter(object):
    def __init__(self, columns):
        self.df = pd.DataFrame(columns=columns)
        self.columns = columns

    def append(self, **row_data):
        if set(self.columns) == set(row_data):
            new_df = pd.DataFrame.from_records([row_data])
            if len(self.df) == 0:
                self.df = new_df
            else:
                self.df = pd.concat([self.df, new_df], ignore_index=True)


def plot_results(df, target_lib="", audio_format="", ext="png"):
    sns.set_style("whitegrid")

    ordered_libs = df.time.groupby(df.lib).mean().sort_values().index.tolist()

    fig = plt.figure()

    g = sns.catplot(
        x="duration",
        y="time",
        kind="point",
        hue_order=ordered_libs,
        hue="lib",
        data=df,
        height=6.6,
        aspect=1,
    )

    g.savefig("benchmark_%s_%s_dur.%s" % (target_lib, audio_format, ext))

    fig = plt.figure()
    sns.barplot(x="time", y="lib", data=df, order=ordered_libs, orient="h")
    fig.savefig(
        f"benchmark_{target_lib}_{audio_format}_bar.{ext}",
        bbox_inches="tight",
    )



import audioread.ffdec
import audioread.gstdec
import audioread.maddec
import audioread.rawread
import librosa
import numpy as np
import pedalboard
from scipy.io import wavfile
import soundfile as sf
import sox

import audiofile as af


"""
Some of the code taken from:
https://github.com/aubio/aubio/blob/master/python/demos/demo_reading_speed.py
"""


def load_soundfile(fp):
    sig, rate = sf.read(fp)
    return sig


def load_scipy(fp):
    rate, sig = wavfile.read(fp)
    sig = sig.astype("float32") / 32767
    return sig


def load_scipy_mmap(fp):
    rate, sig = wavfile.read(fp, mmap=True)
    sig = sig.astype("float32") / 32767
    return sig


def load_ar_gstreamer(fp):
    with audioread.gstdec.GstAudioFile(fp) as f:
        total_frames = 0
        for buf in f:
            sig = _convert_buffer_to_float(buf)
            sig = sig.reshape(f.channels, -1)
            total_frames += sig.shape[1]
        return sig


def load_ar_mad(fp):
    with audioread.maddec.MadAudioFile(fp) as f:
        total_frames = 0
        for buf in f:
            sig = _convert_buffer_to_float(buf)
            sig = sig.reshape(f.channels, -1)
            total_frames += sig.shape[1]
        return sig


def load_ar_ffmpeg(fp):
    with audioread.ffdec.FFmpegAudioFile(fp) as f:
        total_frames = 0
        for buf in f:
            sig = _convert_buffer_to_float(buf)
            sig = sig.reshape(f.channels, -1)
            total_frames += sig.shape[1]
        return sig


def load_librosa(fp):
    # loading with `sr=None` is disabling the internal resampling
    sig, rate = librosa.load(fp, sr=None)
    return sig


def load_audiofile(fp):
    sig, rate = af.read(fp)
    return sig


def load_pedalboard(fp):
    with pedalboard.io.AudioFile(fp) as f:
        # Pedalboard output is (num_channels, num_samples)
        return f.read(f.frames)[0]


def _convert_buffer_to_float(buf, n_bytes=2, dtype=np.float32):
    # taken from librosa.util.utils
    # Invert the scale of the data
    scale = 1.0 / float(1 << ((8 * n_bytes) - 1))
    # Construct the format string
    fmt = f"<i{n_bytes:d}"
    # Rescale and format the data buffer
    out = scale * np.frombuffer(buf, fmt).astype(dtype)
    return out


def info_soundfile(fp):
    info = {}
    info["duration"] = sf.info(fp).duration
    info["samples"] = int(sf.info(fp).duration * sf.info(fp).samplerate)
    info["channels"] = sf.info(fp).channels
    info["sampling_rate"] = sf.info(fp).samplerate
    return info


def info_audioread(fp):
    info = {}
    with audioread.audio_open(fp) as f:
        info["duration"] = f.duration
    with audioread.audio_open(fp) as f:
        info["samples"] = int(f.duration * f.samplerate)
    with audioread.audio_open(fp) as f:
        info["channels"] = f.channels
    with audioread.audio_open(fp) as f:
        info["sampling_rate"] = f.samplerate
    return info


def info_ar_mad(fp):
    info = {}
    with audioread.maddec.MadAudioFile(fp) as f:
        info["duration"] = f.duration
    with audioread.maddec.MadAudioFile(fp) as f:
        info["samples"] = int(f.duration * f.samplerate)
    with audioread.maddec.MadAudioFile(fp) as f:
        info["channels"] = f.channels
    with audioread.maddec.MadAudioFile(fp) as f:
        info["sampling_rate"] = f.samplerate
    return info


def info_ar_ffmpeg(fp):
    info = {}
    with audioread.ffdec.FFmpegAudioFile(fp) as f:
        info["duration"] = f.duration
    with audioread.ffdec.FFmpegAudioFile(fp) as f:
        info["samples"] = int(f.duration * f.samplerate)
    with audioread.ffdec.FFmpegAudioFile(fp) as f:
        info["channels"] = f.channels
    with audioread.ffdec.FFmpegAudioFile(fp) as f:
        info["sampling_rate"] = f.samplerate
    return info


def info_sox(fp):
    info = {}
    info["duration"] = sox.file_info.duration(fp)
    info["samples"] = sox.file_info.num_samples(fp)
    info["channels"] = sox.file_info.channels(fp)
    info["sampling_rate"] = int(sox.file_info.sample_rate(fp))
    return info


def info_audiofile(fp):
    info = {}
    info["duration"] = af.duration(fp)
    info["samples"] = af.samples(fp)
    info["channels"] = af.channels
    info["sampling_rate"] = af.sampling_rate
    return info


def info_audiofile_sloppy(fp):
    info = {}
    info["duration"] = af.duration(fp, sloppy=True)
    info["samples"] = af.samples(fp)
    info["channels"] = af.channels
    info["sampling_rate"] = af.sampling_rate
    return info


def info_pedalboard(fp):
    info = {}
    with pedalboard.io.AudioFile(fp) as af:
        info["duration"] = af.duration
    with pedalboard.io.AudioFile(fp) as af:
        info["samples"] = af.frames
    with pedalboard.io.AudioFile(fp) as af:
        info["channels"] = af.num_channels
    with pedalboard.io.AudioFile(fp) as af:
        info["sampling_rate"] = af.samplerate
    return info



import argparse
import os
import os.path
import time

import loaders
import utils


def get_files(dir, extension):
    audio_files = []
    dir = os.path.expanduser(dir)
    for root, _, fnames in sorted(os.walk(dir)):
        for fname in fnames:
            if fname.endswith(extension):
                path = os.path.join(root, fname)
                item = path
                audio_files.append(item)
    return audio_files


class AudioFolder(object):
    def __init__(
        self,
        root,
        download=True,
        extension="wav",
        lib="librosa",
    ):
        self.root = os.path.expanduser(root)
        self.data = []
        self.audio_files = get_files(dir=self.root, extension=extension)
        self.loader_function = getattr(loaders, lib)

    def __getitem__(self, index):
        return self.loader_function(self.audio_files[index])

    def __len__(self):
        return len(self.audio_files)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Process some integers.")
    parser.add_argument("--ext", type=str, default="wav")
    args = parser.parse_args()

    columns = [
        "ext",
        "lib",
        "duration",
        "time",
    ]

    store = utils.DFWriter(columns)

    # libraries to be benchmarked
    libs = [
        "ar_ffmpeg",
        "ar_mad",
        "audiofile",
        "audiofile_sloppy",
        "pedalboard",
        "soundfile",
        "sox",
    ]

    audio_walk = sorted(os.walk("AUDIO"))
    if len(audio_walk) == 0:
        raise RuntimeError(
            "No audio files were found.\n"
            "Make sure you executed 'bash generate_audio.sh'"
        )

    for lib in libs:
        # MP4 and MP3 is not supported by all libraries
        if lib in ["soundfile", "sox", "pedalboard"] and args.ext == "mp4":
            continue
        if lib == "ar_ffmpeg" and args.ext == "mp3":  # too slow
            continue
        if lib == "ar_mad" and args.ext != "mp3":
            continue

        print(f"Benchmark metadata {args.ext} with {lib}")
        for root, dirs, fnames in audio_walk:
            for audio_dir in dirs:
                duration = int(audio_dir)
                dataset = AudioFolder(
                    os.path.join(root, audio_dir),
                    lib="info_" + lib,
                    extension=args.ext,
                )

                start = time.time()

                # with contextlib.suppress(UserWarning):
                for fp in dataset.audio_files:
                    info = dataset.loader_function(fp)
                    info["duration"]

                end = time.time()

                # Store ar_ffmpeg and ar_mad as audioread
                if lib in ["ar_ffmpeg", "ar_mad"]:
                    lib_name = "audioread"
                else:
                    lib_name = lib

                store.append(
                    ext=args.ext,
                    lib=lib_name,
                    duration=duration,
                    time=float(end - start) / len(dataset),
                )

    store.df.to_pickle(f"results/benchmark_info_{args.ext}.pickle")



import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns


NBFILES = 10  # must be identical to generate_audio.sh

MAPPINGS = {  # library name mappings
    "audiofile": "audiofile",
    "audiofile_sloppy": "audiofile (sloppy)",
    "audioread": "audioread",
    "librosa": "librosa",
    "pedalboard": "pedalboard",
    "scipy": "scipy",
    "soundfile": "soundfile",
    "sox": "sox",
}

for package in ["read", "info"]:
    dfs = []
    for ext in ["wav", "mp3", "mp4", "ogg", "flac"]:
        dfs.append(pd.read_pickle(f"results/benchmark_{package}_{ext}.pickle"))

    df = pd.concat(dfs, ignore_index=True)

    sns.set_style("whitegrid")

    df_ = df[df["ext"] == "wav"]
    df_ = df_[df_["lib"] == "audiofile"]
    number_of_files = len(df_) * NBFILES

    df["lib"] = df["lib"].map(MAPPINGS)

    if package == "read":
        extensions = [["wav", "flac", "ogg", "mp3", "mp4"]]  # single graph
    else:
        extensions = [["wav", "flac", "ogg", "mp3"], ["mp4"]]

    for exts in extensions:
        y = df[df["ext"].isin(exts)]

        # Define what to show in each figure
        if package == "read":
            lib_order = [
                MAPPINGS["audiofile"],
                MAPPINGS["soundfile"],
                MAPPINGS["librosa"],
                MAPPINGS["pedalboard"],
                MAPPINGS["audioread"],
                MAPPINGS["scipy"],
            ]
            height = 5.8
            aspect = 1.0
        elif package == "info" and "wav" in exts:
            lib_order = [
                MAPPINGS["audiofile"],
                MAPPINGS["soundfile"],
                MAPPINGS["pedalboard"],
                MAPPINGS["audioread"],
            ]
            # Remove audioread for WAV, FLAC, OGG
            y = y[~((y["ext"] != "mp3") & (y["lib"] == "audioread"))]
            height = 4.0
            aspect = 1.6
        elif package == "info" and "mp4" in exts:
            lib_order = [
                MAPPINGS["audiofile"],
                MAPPINGS["audiofile_sloppy"],
                MAPPINGS["audioread"],
            ]
            height = 1.4
            aspect = 4.8

        fig = plt.figure()

        # Define colors for the libraries
        palette = {
            MAPPINGS["audiofile"]: "#4a74b5",
            MAPPINGS["audiofile_sloppy"]: "#6b93d7",
            MAPPINGS["soundfile"]: "#db8548",
            MAPPINGS["librosa"]: "#c34c4d",
            MAPPINGS["scipy"]: "#8174b8",
            MAPPINGS["audioread"]: "#94785e",
            MAPPINGS["sox"]: "#db8cc5",
            MAPPINGS["pedalboard"]: "#5dab64",
        }

        g = sns.catplot(
            x="time",
            y="ext",
            kind="bar",
            hue="lib",
            hue_order=lib_order,
            order=exts,
            palette=palette,
            data=y,
            height=height,
            aspect=aspect,
            legend=True,
        )
        g.despine(left=True)
        plt.xlabel("time / s per file")
        plt.ylabel("file format")
        if package == "info":
            plt.title(f"Access metadata, average over {number_of_files} files")
        else:
            plt.title(f"Read file, average over {number_of_files} files")
        g.savefig(f'results/benchmark_{"-".join(exts)}_{package}.png')
        plt.close()



from __future__ import annotations

import os
import tempfile

import numpy as np
import soundfile

import audeer
import audmath

from audiofile.core.convert import convert
from audiofile.core.utils import MAX_CHANNELS
from audiofile.core.utils import SNDFORMATS
from audiofile.core.utils import duration_in_seconds
from audiofile.core.utils import file_extension


def convert_to_wav(
    infile: str,
    outfile: str = None,
    offset: float | int | str | np.timedelta64 = None,
    duration: float | int | str | np.timedelta64 = None,
    bit_depth: int = 16,
    normalize: bool = False,
    overwrite: bool = False,
    **kwargs,
) -> str:
    """Convert any audio/video file to WAV.

    It uses soundfile for reading WAV, FLAC, MP3, OGG files,
    and sox or ffmpeg for reading all other files.
    If ``duration`` and/or ``offset`` are specified
    the resulting WAV file
    will be shortened accordingly.

    ``duration`` and ``offset``
    support all formats
    mentioned in :func:`audmath.duration_in_seconds`,
    like ``'2 ms'``, or ``pd.to_timedelta(2, 's')``.
    The exception is
    that float and integer values
    are always interpreted as seconds
    and strings without unit
    always as samples.
    If ``duration`` and/or ``offset`` are negative,
    they are interpreted from right to left,
    whereas ``duration`` starts from the end of the signal
    for ``offset=None``.
    If the signal is shorter than the requested ``duration`` and/or ``offset``
    only the part of the signal overlapping with the requested signal
    is returned,
    e.g. for a file containing the signal ``[0, 1, 2]``,
    ``duration=2``, ``offset=-4`` will return ``[0]``.

    ``duration`` and ``offset``
    are evenly rounded
    after conversion to samples.

    It then uses :func:`soundfile.write` to write the WAV file,
    which limits the number of supported channels to 65535.

    Args:
        infile: audio/video file name
        outfile: WAV file name.
            If ``None`` same path as ``infile``
            but file extension is replaced by ``'wav'``
        duration: return only a specified duration
        offset: start reading at offset
        bit_depth: bit depth of written file in bit,
            can be 8, 16, 24
        normalize: normalize audio data before writing
        overwrite: force overwriting
            if ``outfile`` is identical to ``outfile``
        kwargs: pass on further arguments to :func:`soundfile.write`

    Returns:
        absolute path to resulting WAV file

    Raises:
        FileNotFoundError: if ffmpeg binary is needed,
            but cannot be found
        RuntimeError: if ``file`` is missing,
            broken or format is not supported
        RuntimeError: if ``infile`` would need to be overwritten
            and ``overwrite`` is ``False``
        ValueError: if ``duration`` is a string
            that does not match a valid '<value><unit>' pattern
            or the provided unit is not supported

    Examples:
        >>> path = convert_to_wav("stereo.flac")
        >>> os.path.basename(path)
        'stereo.wav'

    """
    infile = audeer.safe_path(infile)
    if outfile is None:
        outfile = audeer.replace_file_extension(infile, "wav")
    else:
        outfile = audeer.safe_path(outfile)
    if infile == outfile and not overwrite:
        raise RuntimeError(
            f"'{infile}' would be overwritten. "
            "Select 'overwrite=True', "
            "or provide an 'outfile' argument."
        )
    signal, sampling_rate = read(
        infile,
        offset=offset,
        duration=duration,
    )
    write(
        outfile,
        signal,
        sampling_rate,
        bit_depth=bit_depth,
        normalize=normalize,
        **kwargs,
    )
    return outfile


def read(
    file: str,
    duration: float | int | str | np.timedelta64 = None,
    offset: float | int | str | np.timedelta64 = None,
    always_2d: bool = False,
    dtype: str = "float32",
    **kwargs,
) -> tuple[np.array, int]:
    """Read audio file.

    It uses :func:`soundfile.read` for WAV, FLAC, MP3, and OGG files.
    All other audio files are
    first converted to WAV by sox or ffmpeg.

    ``duration`` and ``offset``
    support all formats
    mentioned in :func:`audmath.duration_in_seconds`,
    like ``'2 ms'``, or ``pd.to_timedelta(2, 's')``.
    The exception is
    that float and integer values
    are always interpreted as seconds
    and strings without unit
    always as samples.
    If ``duration`` and/or ``offset`` are negative,
    they are interpreted from right to left,
    whereas ``duration`` starts from the end of the signal
    for ``offset=None``.
    If the signal is shorter than the requested ``duration`` and/or ``offset``
    only the part of the signal overlapping with the requested signal
    is returned,
    e.g. for a file containing the signal ``[0, 1, 2]``,
    ``duration=2``, ``offset=-4`` will return ``[0]``.

    ``duration`` and ``offset``
    are evenly rounded
    after conversion to samples.

    Args:
        file: file name of input audio file
        duration: return only the specified duration
        offset: start reading at offset
        always_2d: if ``True`` it always returns a two-dimensional signal
            even for mono sound files
        dtype: data type of returned signal,
            select from
            ``'float64'``,
            ``'float32'``,
            ``'int32'``,
            ``'int16'``
        kwargs: pass on further arguments to :func:`soundfile.read`

    Returns:
        * a two-dimensional array in the form
          ``[channels, samples]``.
          If the sound file has only one channel
          and ``always_2d=False``,
          a one-dimensional array is returned
        * sample rate of the audio file

    Raises:
        FileNotFoundError: if ffmpeg binary is needed,
            but cannot be found
        RuntimeError: if ``file`` is missing,
            broken or format is not supported
        ValueError: if ``duration`` is a string
            that does not match a valid '<value><unit>' pattern
            or the provided unit is not supported

    Examples:
        .. plot::
            :context: reset
            :include-source: false

            import numpy as np
            from audiofile.core.io import read

        .. plot::
            :context: close-figs

            >>> signal, sampling_rate = read("mono.wav", always_2d=True)
            >>> sampling_rate
            8000
            >>> signal.shape
            (1, 12000)
            >>> signal, sampling_rate = read("mono.wav")
            >>> signal.shape
            (12000,)
            >>> import audplot
            >>> audplot.waveform(signal)

        .. plot::
            :context: close-figs

            >>> signal, sampling_rate = read("mono.wav", duration=0.5)
            >>> # Extend signal to original length
            >>> signal = np.pad(signal, (0, 8000))
            >>> audplot.waveform(signal)

        .. plot::
            :context: close-figs

            >>> signal, sampling_rate = read("mono.wav", duration=-0.5)
            >>> # Extend signal to original length
            >>> signal = np.pad(signal, (8000, 0))
            >>> audplot.waveform(signal)

        .. plot::
            :context: close-figs

            >>> signal, sampling_rate = read("mono.wav", offset="4000", duration="4000")
            >>> # Extend signal to original length
            >>> signal = np.pad(signal, (4000, 4000))
            >>> audplot.waveform(signal)

        .. plot::
            :context: close-figs

            >>> # Use audresample for resampling and remixing
            >>> import audresample
            >>> signal, sampling_rate = read("stereo.wav")
            >>> signal.shape
            (2, 12000)
            >>> target_rate = 16000
            >>> signal = audresample.resample(signal, sampling_rate, target_rate)
            >>> signal.shape
            (2, 24000)
            >>> signal = audresample.remix(signal, mixdown=True)
            >>> signal.shape
            (1, 24000)
            >>> audplot.waveform(signal)

    """  # noqa: E501
    file = audeer.safe_path(file)
    sampling_rate = None

    # Parse offset and duration values
    if (
        duration is not None
        or isinstance(duration, str)
        or (offset is not None and isinstance(offset, str))
        or (offset is not None and offset != 0)
    ):
        # Import sampling_rate here to avoid circular imports
        from audiofile.core.info import sampling_rate as get_sampling_rate

        sampling_rate = get_sampling_rate(file)
    if duration is not None:
        duration = duration_in_seconds(duration, sampling_rate)
        if np.isnan(duration):
            duration = None
    if offset is not None and offset != 0:
        offset = duration_in_seconds(offset, sampling_rate)
        if np.isnan(offset):
            offset = None

    # Support for negative offset/duration values
    # by counting them from end of signal
    #
    if offset is not None and offset < 0 or duration is not None and duration < 0:
        # Import duration here to avoid circular imports
        from audiofile.core.info import duration as get_duration

        signal_duration = get_duration(file)
    # offset | duration
    # None   | < 0
    if offset is None and duration is not None and duration < 0:
        offset = max([0, signal_duration + duration])
        duration = None
    # None   | >= 0
    if offset is None and duration is not None and duration >= 0:
        if np.isinf(duration):
            duration = None
    # >= 0   | < 0
    elif offset is not None and offset >= 0 and duration is not None and duration < 0:
        if np.isinf(offset) and np.isinf(duration):
            offset = 0
            duration = None
        elif np.isinf(offset):
            duration = 0
        else:
            if np.isinf(duration):
                offset = min([offset, signal_duration])
                duration = np.sign(duration) * signal_duration
            orig_offset = offset
            offset = max([0, offset + duration])
            duration = min([-duration, orig_offset])
    # >= 0   | >= 0
    elif offset is not None and offset >= 0 and duration is not None and duration >= 0:
        if np.isinf(offset):
            duration = 0
        elif np.isinf(duration):
            duration = None
    # < 0    | None
    elif offset is not None and offset < 0 and duration is None:
        offset = max([0, signal_duration + offset])
    # >= 0    | None
    elif offset is not None and offset >= 0 and duration is None:
        if np.isinf(offset):
            duration = 0
    # < 0    | > 0
    elif offset is not None and offset < 0 and duration is not None and duration > 0:
        if np.isinf(offset) and np.isinf(duration):
            offset = 0
            duration = None
        elif np.isinf(offset):
            duration = 0
        elif np.isinf(duration):
            duration = None
        else:
            offset = signal_duration + offset
            if offset < 0:
                duration = max([0, duration + offset])
            else:
                duration = min([duration, signal_duration - offset])
            offset = max([0, offset])
    # < 0    | < 0
    elif offset is not None and offset < 0 and duration is not None and duration < 0:
        if np.isinf(offset):
            duration = 0
        elif np.isinf(duration):
            duration = -signal_duration
        else:
            orig_offset = offset
            offset = max([0, signal_duration + offset + duration])
            duration = min([-duration, signal_duration + orig_offset])
            duration = max([0, duration])

    # Convert to samples
    #
    # Handle duration first
    # and returned immediately
    # if duration == 0
    if duration is not None and duration != 0:
        duration = audmath.samples(duration, sampling_rate)
    if duration == 0:
        from audiofile.core.info import channels as get_channels

        channels = get_channels(file)
        if channels > 1 or always_2d:
            signal = np.zeros((channels, 0))
        else:
            signal = np.zeros((0,))
        return signal, sampling_rate
    if offset is not None and offset != 0:
        offset = audmath.samples(offset, sampling_rate)
    else:
        offset = 0

    tmpdir = None
    if file_extension(file) not in SNDFORMATS:
        # Convert file formats not recognized by soundfile to WAV first.
        #
        # NOTE: this is faster than loading them with librosa directly.
        # In addition, librosa seems to have an issue with the precision of
        # the returned magnitude
        # (https://github.com/librosa/librosa/issues/811).
        #
        with tempfile.TemporaryDirectory(prefix="audiofile") as tmpdir:
            tmpfile = os.path.join(tmpdir, "tmp.wav")
            # offset and duration have to be given in seconds
            if offset != 0:
                offset /= sampling_rate
            if duration is not None and duration != 0:
                duration /= sampling_rate
            if sampling_rate is None:
                # Infer sampling rate using mediainfo before conversion,
                # as ffmpeg does ignore the original sampling rate for opus files,
                # see:
                # * https://trac.ffmpeg.org/ticket/5240
                # * https://github.com/audeering/audiofile/issues/157
                from audiofile.core.info import sampling_rate as get_sampling_rate

                sampling_rate = get_sampling_rate(file)
            convert(file, tmpfile, offset, duration, sampling_rate)
            signal, sampling_rate = soundfile.read(
                tmpfile,
                dtype=dtype,
                always_2d=always_2d,
                **kwargs,
            )
    else:
        start = offset
        # duration == 0 is handled further above with immediate return
        if duration is not None:
            stop = duration + start
        else:
            stop = None
        signal, sampling_rate = soundfile.read(
            file,
            start=start,
            stop=stop,
            dtype=dtype,
            always_2d=always_2d,
            **kwargs,
        )
    # [samples, channels] => [channels, samples]
    signal = signal.T
    return signal, sampling_rate


def write(
    file: str,
    signal: np.array,
    sampling_rate: int,
    bit_depth: int = 16,
    normalize: bool = False,
    **kwargs,
):
    """Write (normalized) audio files.

    Save audio data provided as an array of shape ``(channels, samples)``
    or ``(samples,)``
    to a WAV, FLAC, NP3, or OGG file.
    ``channels`` can be up to 65535 for WAV,
    255 for OGG,
    2 for MP3,
    and 8 for FLAC.
    For monaural audio the array can be one-dimensional.

    It uses :func:`soundfile.write` to write the audio files.

    Args:
        file: file name of output audio file.
            The format (WAV, FLAC, MP3, OGG) will be inferred from the file name
        signal: audio data to write
        sampling_rate: sample rate of the audio data
        bit_depth: bit depth of written file in bit,
            can be 8, 16, 24 for WAV and FLAC files,
            and in addition 32 for WAV files
        normalize: normalize audio data before writing
        kwargs: pass on further arguments to :func:`soundfile.write`

    Raises:
        RuntimeError: for non-supported bit depth or number of channels

    Examples:
        >>> sampling_rate = 8000
        >>> signal = np.random.uniform(-1, 1, (1, 1000))
        >>> write("mono.wav", signal, sampling_rate)
        >>> signal = np.random.uniform(-1.2, 1.2, (2, 1000))
        >>> write("stereo.flac", signal, sampling_rate, normalize=True)

    """  # noqa: E501
    file = audeer.safe_path(file)
    file_type = file_extension(file)

    # Check for allowed precisions
    if file_type == "wav":
        depth_mapping = {
            8: "PCM_U8",
            16: "PCM_16",
            24: "PCM_24",
            32: "PCM_32",
        }
    elif file_type == "flac":
        depth_mapping = {
            8: "PCM_S8",
            16: "PCM_16",
            24: "PCM_24",
        }
    if file_type in ["wav", "flac"]:
        bit_depths = sorted(list(depth_mapping.keys()))
        if bit_depth not in bit_depths:
            raise RuntimeError(
                f'"bit_depth" has to be one of '
                f'{", ".join([str(b) for b in bit_depths])}.'
            )
        subtype = depth_mapping[bit_depth]
    else:
        subtype = None
    # Check if number of channels is allowed for chosen file type
    if signal.ndim > 1:
        channels = np.shape(signal)[0]
    else:
        channels = 1
    if channels > MAX_CHANNELS[file_type]:
        if file_type != "wav":
            hint = " Consider using 'wav' instead."
        else:
            hint = ""
        raise RuntimeError(
            "The maximum number of allowed channels "
            f"for '{file_type}' is {MAX_CHANNELS[file_type]}.{hint}"
        )
    if normalize:
        signal = signal / np.max(np.abs(signal))
    soundfile.write(file, signal.T, sampling_rate, subtype=subtype, **kwargs)



import subprocess

from audiofile.core.utils import binary_missing_error
from audiofile.core.utils import broken_file_error
from audiofile.core.utils import run_ffmpeg
from audiofile.core.utils import run_sox


def convert(
    infile: str,
    outfile: str,
    offset: float = 0,
    duration: float = None,
    sampling_rate: int = None,
):
    """Convert any audio/video file to WAV.

    Args:
        infile: audio/video file name
        outfile: WAV file name
        duration: return only a specified duration in seconds
        offset: start reading at offset in seconds
        sampling_rate: sampling rate in Hz

    """
    try:
        # Convert to WAV file with sox
        run_sox(infile, outfile, offset, duration, sampling_rate)
    except (FileNotFoundError, subprocess.CalledProcessError):
        try:
            # Convert to WAV file with ffmpeg
            run_ffmpeg(infile, outfile, offset, duration, sampling_rate)
        except FileNotFoundError as e:  # pragma: no cover
            raise binary_missing_error("ffmpeg") from e
        except subprocess.CalledProcessError as e:  # pragma: no cover
            raise broken_file_error(infile) from e



import os

import numpy as np
import pytest

import audeer

import audiofile


np.random.seed(1)


@pytest.fixture(scope="session", autouse=True)
def audio_file():
    create_audio_files(".")

    yield

    # Clean up
    for file in audeer.list_file_names(".", filetype="wav"):
        if os.path.exists(file):
            os.remove(file)
    for file in audeer.list_file_names(".", filetype="flac"):
        if os.path.exists(file):
            os.remove(file)


def create_audio_files(
    basedir,
):
    sampling_rate = 8000
    mono_wav_file = audeer.path(basedir, "mono.wav")
    stereo_wav_file = audeer.path(basedir, "stereo.wav")
    stereo_flac_file = audeer.path(basedir, "stereo.flac")
    signal = am_fm_synth(1.5, 1, sampling_rate)
    audiofile.write(mono_wav_file, signal, sampling_rate)
    signal = am_fm_synth(1.5, 2, sampling_rate)
    audiofile.write(stereo_wav_file, signal, sampling_rate)
    audiofile.write(stereo_flac_file, signal, sampling_rate)


def am_fm_synth(
    duration: float,
    num_channels: int = 1,
    sampling_rate: int = 16000,
    *,
    dtype: type = np.float32,
) -> np.ndarray:
    r"""Synthesise an AM/FM signal of given duration (sampled at given rate).

    Args:
        duration: duration in seconds
        num_channels: number of channels in the output signal
        sampling_rate: sampling rate in Hz
        dtype: data type

    Returns:
        Synthesised signal with shape `(number of channels, number of samples)`

    """
    n_samples = int(duration * sampling_rate)
    g = 0.8  # gain
    g_am = 0.7  # amount of AM
    f_am = 2.5  # frequency of AM (Hz)
    f0 = 0.04 * sampling_rate  # carrier frequency (Hz)
    f_mod = 2  # modulator frequency (Hz)
    f_dev = f0 * 0.95  # frequency deviation (intensity of FM)
    omega_am = 2 * np.pi * f_am / sampling_rate
    omega0_car = 2 * np.pi * f0 / sampling_rate
    omega_mod = 2 * np.pi * f_mod / sampling_rate
    omega_dev = 2 * np.pi * f_dev / sampling_rate
    ph_fm = 0  # initial phase of FM oscillator
    ph_am = np.pi / 2  # initial phase of AM oscillator
    sig = np.zeros((num_channels, n_samples), dtype=dtype)
    for ch_indx in range(num_channels):
        # No reinitialisation (to get true stereo)
        for t in range(n_samples):
            sig[ch_indx, t] = g * np.cos(ph_fm)
            sig[ch_indx, t] *= (1 - g_am) + g_am * np.square(np.cos(ph_am))
            ph_am += omega_am / 2
            ph_fm += omega0_car + omega_dev * np.cos(omega_mod * t)
    return sig
